# -*- coding: utf-8 -*-
"""Created on Wed Sep 12 12:15:42 2018.

@author: Artesia
"""

import os
import tempfile
import time
import zipfile

import numpy as np
import pandas as pd
import scipy.spatial.qhull as qhull
from pandas import Timedelta, Timestamp


def _obslist_to_frame(obs_list):
    """convert a list of observations to a pandas DataFrame.

    Parameters
    ----------
    obs_list : TYPE
        DESCRIPTION.

    Returns
    -------
    obs_df : TYPE
        DESCRIPTION.
    """
    if len(obs_list) > 0:
        obs_df = pd.DataFrame([o.to_collection_dict() for o in obs_list],
                              columns=obs_list[0].to_collection_dict().keys())
        obs_df.set_index('name', inplace=True)
    else:
        obs_df = pd.DataFrame()
    return obs_df


def unzip_file(src, dst, force=False, preserve_datetime=False):
    """Unzip file.

    Parameters
    ----------
    src : str
        source zip file
    dst : str
        destination directory
    force : boolean, optional
        force unpack if dst already exists
    preserve_datetime : boolean, optional
        use date of the zipfile for the destination file

    Returns
    -------
    int
        1 of True
    """
    if os.path.exists(dst):
        if not force:
            print(
                "File not unzipped. Destination already exists. Use 'force=True' to unzip.")
            return
    if preserve_datetime:
        zipf = zipfile.ZipFile(src, 'r')
        for f in zipf.infolist():
            zipf.extract(f, path=dst)
            date_time = time.mktime(f.date_time + (0, 0, -1))
            os.utime(os.path.join(dst, f.filename), (date_time, date_time))
        zipf.close()
    else:
        zipf = zipfile.ZipFile(src, 'r')
        zipf.extractall(dst)
        zipf.close()
    return 1


def unzip_changed_files(zipname, pathname, check_time=True, check_size=False,
                        verbose=False):
    # Extract each file in a zip-file only when the properties are different
    # With the default arguments this method only checks the modification time
    with zipfile.ZipFile(zipname) as zf:
        infolist = zf.infolist()
        for info in infolist:
            fname = os.path.join(pathname, info.filename)
            extract = False
            if os.path.exists(fname):
                if check_time:
                    tz = time.mktime(info.date_time + (0, 0, -1))
                    tf = os.path.getmtime(fname)
                    if tz != tf:
                        extract = True
                if check_size:
                    sz = info.file_size
                    sf = os.path.getsize(fname)
                    if sz != sf:
                        extract = True
            else:
                extract = True
            if extract:
                if verbose:
                    print('extracting {}'.format(info.filename))
                zf.extract(info.filename, pathname)
                # set the correct modification time
                # (which is the time of extraction by default)
                tz = time.mktime(info.date_time + (0, 0, -1))
                os.utime(os.path.join(pathname, info.filename), (tz, tz))


def matlab2datetime(tindex):
    """Transform a matlab time to a datetime, rounded to seconds."""
    day = Timestamp.fromordinal(int(tindex))
    dayfrac = Timedelta(days=float(tindex) % 1) - Timedelta(days=366)
    return day + dayfrac


def get_files(file_or_dir, ext, unpackdir=None, force_unpack=False,
              preserve_datetime=False):
    """internal method to get list of files with specific extension from
    dirname.

    Parameters
    ----------
    file_or_dir : str
        file or path to data
    ext : str
        extension of filenames to store in list
    force_unpack : bool, optional
        force unzip, by default False
    preserve_datetime : bool, optional
        preserve datetime of unzipped files, by default False
        (useful for checking whether data has changed)
    """
    # check if unpackdir is same as file_or_dir, if same, this can cause
    # problems when the unpackdir still contains zips that will be unpacked
    # again.
    if unpackdir is not None:
        if os.path.normcase(unpackdir) == os.path.normcase(file_or_dir):
            raise ValueError("Please specify a different folder to unpack"
                             " files!")
    # identify whether file_or_dir started as zip
    if file_or_dir.endswith(".zip"):
        iszip = True
    else:
        iszip = False

    # unzip dir
    if file_or_dir.endswith('.zip'):
        zipf = file_or_dir
        if unpackdir is None:
            file_or_dir = tempfile.TemporaryDirectory().name
        else:
            file_or_dir = unpackdir
        unzip_file(zipf, file_or_dir, force=force_unpack,
                   preserve_datetime=preserve_datetime)

    # file_or_dir is directory
    if os.path.isdir(file_or_dir):
        # check for zips in dir
        zip_fnames = [i for i in os.listdir(
            file_or_dir) if i.endswith(".zip")]
        if len(zip_fnames) > 0:
            # unzip zips
            if unpackdir is None:
                dirname = tempfile.TemporaryDirectory().name
            else:
                dirname = unpackdir
            for zipf in zip_fnames:
                unzip_file(os.path.join(file_or_dir, zipf), dirname,
                           force=True,
                           preserve_datetime=preserve_datetime)
                # remove intermediate zipfiles if initial file_or_dir was zip
                if iszip:
                    os.remove(os.path.join(file_or_dir, zipf))
        else:
            dirname = file_or_dir
        # get all files with extension ext
        unzip_fnames = [i for i in os.listdir(
            dirname) if i.endswith(ext)]
    elif os.path.isfile(file_or_dir):
        # file_or_dir is actually an xml
        unzip_fnames = [os.path.basename(file_or_dir)]  # get file name
        dirname = os.path.dirname(file_or_dir)  # get directory path
    else:
        raise NotImplementedError("Cannot parse 'file_or_dir': "
                                  f"{file_or_dir}!")

    return dirname, unzip_fnames


def interp_weights(xy, uv, d=2):
    """Calculate interpolation weights [1]_.

    Parameters
    ----------
    xy : np.array
        array containing x-coordinates in first column and y-coordinates
        in second column
    uv : np.array
        array containing coordinates at which interpolation weights should
        be calculated, x-data in first column and y-data in second column
    d : int, optional
        dimension of data? (the default is 2, which works for 2D data)

    Returns
    -------
    vertices: np.array
        array containing interpolation vertices
    weights: np.array
        array containing interpolation weights per point


    References
    ----------
    .. [1] https://stackoverflow.com/questions/20915502/speedup-scipy-griddata-for-multiple-interpolations-between-two-irregular-grids
    """

    tri = qhull.Delaunay(xy)
    simplex = tri.find_simplex(uv)
    vertices = np.take(tri.simplices, simplex, axis=0)
    temp = np.take(tri.transform, simplex, axis=0)
    delta = uv - temp[:, d]
    bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)
    return vertices, np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))


def interpolate(values, vtx, wts):
    """Interpolate values at locations defined by vertices and points [2]_, as
    calculated by interp_weights function.

    Parameters
    ----------
    values : np.array
        array containing values to interpolate
    vtx : np.array
        array containing interpolation vertices, see interp_weights()
    wts : np.array
        array containing interpolation weights, see interp_weights()

    Returns
    -------
    arr: np.array
        array containing interpolated values at locations as given by
        vtx and wts

    References
    ----------
    .. [2] https://stackoverflow.com/questions/20915502/speedup-scipy-griddata-for-multiple-interpolations-between-two-irregular-grids
    """

    return np.einsum('nj,nj->n', np.take(values, vtx), wts)


def df2gdf(df, xcol='x', ycol='y'):
    """Make a GeoDataFrame from a DataFrame, assuming the geometry are
    points."""
    from geopandas import GeoDataFrame
    from shapely.geometry import Point
    gdf = GeoDataFrame(df.copy(), geometry=[Point(
        (s[xcol], s[ycol])) for i, s in df.iterrows()])
    return gdf


def show_versions():
    """Method to print the version of dependencies.

    """
    from pandas import __version__ as pd_version
    from numpy import __version__ as np_version
    from scipy import __version__ as sc_version
    from matplotlib import __version__ as mpl_version
    from sys import version as os_version

    msg = (
        f"Python version    : {os_version}\n"
        f"Numpy version     : {np_version}\n"
        f"Scipy version     : {sc_version}\n"
        f"Pandas version    : {pd_version}\n"
        f"Matplotlib version: {mpl_version}"
    )

    return print(msg)



import datetime as dt
import os
import re
import tempfile
from io import StringIO

import numpy as np
import pandas as pd
import requests

from .. import util


def get_stations(meteo_var='RD'):
    """get knmi stations from json files according to variable.

    Parameters
    ----------
    meteo_var : str, optional
        [description], by default 'RD'

    Returns
    -------
    pandas DataFrame with stations, names and coordinates (Lat/Lon & RD)
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))

    if meteo_var == "RD":
        fname = "../data/knmi_neerslagstation.json"
    else:
        fname = "../data/knmi_meteostation.json"

    stations = pd.read_json(os.path.join(dir_path, fname))

    if meteo_var == 'PG':
        # in Ell wordt geen luchtdruk gemeten
        stations.drop(377, inplace=True)
    if meteo_var == 'EV24':
        # in Woensdrecht wordt geen verdamping gemeten
        stations.drop(340, inplace=True)

    return stations


def get_station_name(stn, stations):
    """Returns the station name from a KNMI station.

    Modifies the station name in such a way that a valid url can be obtained.

    Parameters
    ----------
    stn : int
        station number.
    stations : pandas DataFrame
        DataFrame with the station metadata.

    Returns
    -------
    str
        Name of the station.
    """
    stn_name = stations.loc[stn, 'naam']
    stn_name = stn_name.upper().replace(' ', '-')
    stn_name = stn_name.replace('(', '').replace(')', '')

    return stn_name


def get_nearest_stations_xy(x, y, meteo_var, n=1, stations=None, ignore=None):
    """find the KNMI stations that measure 'variable' closest to the x, y
    coordinates.

    Parameters
    ----------
    x : int or float
        x coordinate in RD
    y : int or float
        x coordinate in RD
    meteo_var : str
        measurement variable e.g. 'RD' or 'EV24'
    n : int, optional
        number of stations you want to return. The default is 1.
    stations : pd.DataFrame, optional
        if None stations will be obtained using the get_stations function.
        The default is None.
    ignore : list, optional
        list of stations to ignore. The default is None.

    Returns
    -------
    list
        station numbers.
    """

    if stations is None:
        stations = get_stations(meteo_var=meteo_var)
    if ignore is not None:
        stations.drop(ignore, inplace=True)
        if stations.empty:
            return None

    d = np.sqrt((stations.x - x)**2 + (stations.y - y)**2)

    return d.nsmallest(n).index.to_list()


def get_nearest_station_df(locations, xcol='x', ycol='y', stations=None,
                           meteo_var="RD", ignore=None):
    """find the KNMI stations that measure 'meteo_var' closest to the
    coordinates in 'locations'.

    Parameters
    ----------
    locations : pd.DataFrame
        x and y coordinates
    xcol : str
        name of the column in the locations dataframe with the x values
    ycol : str
        name of teh column in the locations dataframe with the y values
    stations : pd.DataFrame, optional
        if None stations will be obtained using the get_stations function.
        The default is None.
    meteo_var : str
        measurement variable e.g. 'RD' or 'EV24'
    ignore : list, optional
        list of stations to ignore. The default is None.

    Returns
    -------
    stns : list
        station numbers.
    """
    if stations is None:
        stations = get_stations(meteo_var=meteo_var)
    if ignore is not None:
        stations.drop(ignore, inplace=True)
        if stations.empty:
            return None

    xo = pd.to_numeric(locations[xcol])
    xt = pd.to_numeric(stations.x)
    yo = pd.to_numeric(locations[ycol])
    yt = pd.to_numeric(stations.y)

    xh, xi = np.meshgrid(xt, xo)
    yh, yi = np.meshgrid(yt, yo)

    distances = pd.DataFrame(np.sqrt((xh - xi) ** 2 + (yh - yi) ** 2),
                             index=locations.index,
                             columns=stations.index)

    stns = distances.idxmin(axis=1).unique()
    if np.any(np.isnan(stns)):
        stns = stns[~np.isnan(stns)].astype(int)

    return stns


def get_nearest_station_grid(xmid, ymid, stations=None, meteo_var="RD",
                             ignore=None):
    """find the KNMI stations that measure 'meteo_var' closest to all cells in
    a grid.

    Parameters
    ----------
    xmid : np.array
        x coördinates of the cell centers of your grid shape(ncol)
    ymid : np.array
        y coördinates of the cell centers of your grid shape(nrow)
    stations : pd.DataFrame, optional
        if None stations will be obtained using the get_stations function.
        The default is None.
    meteo_var : str
        measurement variable e.g. 'RD' or 'EV24'
    ignore : list, optional
        list of stations to ignore. The default is None.

    Returns
    -------
    stns : list
        station numbers.

    Notes
    -----
    assumes you have a structured rectangular grid.
    """
    mg = np.meshgrid(xmid, ymid)

    locations = pd.DataFrame(data={'x': mg[0].ravel(),
                                   'y': mg[1].ravel()})

    stns = get_nearest_station_df(locations, xcol='x', ycol='y',
                                  stations=stations, meteo_var=meteo_var,
                                  ignore=ignore)

    return stns


def _start_end_to_datetime(start, end):
    """convert start and endtime to datetime.

    Parameters
    ----------
    start : str, datetime, None
        start time
    end : str, datetime, None
        start time

    Returns
    -------
    start : pd.TimeStamp
        start time
    end : pd.TimeStamp
        end time
    """

    if start is None:
        start = pd.Timestamp(pd.Timestamp.today().year - 1, 1, 1)
    else:
        start = pd.to_datetime(start)
    # start date one day before because later the datetime index is modified
    start = start - pd.Timedelta(1, 'D')
    if end is None:
        end = pd.Timestamp.today() - pd.Timedelta(1, unit='D')
    else:
        end = pd.to_datetime(end)

    return start, end


def _check_latest_measurement_date_RD_debilt(use_api=False,
                                             verbose=False):
    """According to the website of the knmi it can take up to 3 weeks before
    precipitation data is updated. If you use the fill_missing_measurements
    method to fill a time series untill today, it will keep looking at all
    stations to find the data of these last days that probably does not exist.

    To prevent this, this function will find the last day there are measure-
    ments at knmi station de Bilt. It is assumed that if de Bilt doesn't have
    recent measurements, no station will have measurements for these dates.

    website knmi: https://www.knmi.nl/nederland-nu/klimatologie/monv/reeksen

    Parameters
    ----------
    use_api : bool, optional
        if True the api is used to obtain the data, API documentation is here:
            https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script
        Default is False as the API is down (apr 2021).
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    last_measurement_date_debilt : pd.TimeStamp
        last date with measurements at station de Bilt
    """

    look_back_days = 90
    start = dt.datetime.now() - pd.Timedelta(look_back_days, unit='D')
    end = dt.datetime.now() + pd.Timedelta(10, unit='D')
    if use_api:
        url = 'http://projects.knmi.nl/klimatologie/monv/reeksen/getdata_rr.cgi'
        knmi_df, _ = get_knmi_daily_rainfall_api(url, 550, "RD", start=start,
                                                 end=end, inseason=False,
                                                 verbose=verbose)
    else:
        knmi_df, _ = get_knmi_daily_rainfall_url(
            550, 'DE-BILT', 'RD', start, end, inseason=False, verbose=verbose)

    knmi_df = knmi_df.dropna()
    if knmi_df.empty:
        raise ValueError(
            f'knmi station de Bilt has no RD measurements in the past {look_back_days} days.')

    last_measurement_date_debilt = knmi_df.index[-1]

    if verbose:
        print(f'last measurement available at the Bilt is from'
              f' {last_measurement_date_debilt.strftime("%Y-%m-%d")}'
              f' changing end_date accordingly')

    return last_measurement_date_debilt


def download_knmi_data(stn, stn_name=None,
                       meteo_var='RD', start=None, end=None,
                       interval='daily', inseason=False,
                       use_api=False,
                       raise_exceptions=True, verbose=False):
    """download knmi data of a measurements station for certain observation
    type.

    Parameters
    ----------
    stn : int or str
        number of measurements
    stn_name : str
        Only necesary if meteo_var is RD and use_api is True.
    meteo_var : str, optional
        measurement type 'RD' or 'EV24'. The default is 'RD'.
    start : str, datetime or None, optional
        start date of observations. The default is None.
    end : str, datetime or None, optional
        end date of observations. The default is None.
    interval : str, optional
        time interval of observations. The default is 'daily'.
    inseason : bool, optional
        passed to the knmi api. The default is False.
    use_api : bool, optional
        if True the api is used to obtain the data, API documentation is here:
            https://www.knmi.nl/kennis-en-datacentrum/achtergrond/data-ophalen-vanuit-een-script
        Default is False as the API is down (apr 2021).
    raise_exceptions : bool, optional
        if True you get errors when no data is returned. The default is True.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Raises
    ------
    NotImplementedError
        different time intervals and inseason data is not yet working.
    ValueError
        if the data from knmi cannot not be read a ValueError is raised.
        Unless raise_exceptions is False

    Returns
    -------
    knmi_df : pd.DataFrame
        data from one station from one type of observation
    variables : dictionary
        information about the observerd variables
    stations : pd.DataFrame
        information about the measurement station.
    """
    # checks
    if interval.startswith('hour') and meteo_var == 'RD':
        message = 'Interval cannot be hourly for rainfall-stations'
        raise (ValueError(message))

    if interval != 'daily':
        raise NotImplementedError('only daily intervals are working now')

    if inseason:
        raise NotImplementedError('season stuff not implemented')

    start, end = _start_end_to_datetime(start, end)

    # convert possible integer to string
    stn = str(stn)

    # define variables
    knmi_df = pd.DataFrame()
    variables = {}
    stations = pd.DataFrame()

    # download and read data
    try:
        if use_api:
            if interval.startswith('hour'):
                # hourly data from meteorological stations
                url = 'http://projects.knmi.nl/klimatologie/uurgegevens/getdata_uur.cgi'
                knmi_df = get_knmi_hourly_api(url, stn, meteo_var, start, end)

            elif meteo_var == 'RD':
                # daily data from rainfall-stations
                url = 'http://projects.knmi.nl/klimatologie/monv/reeksen/getdata_rr.cgi'
                knmi_df, variables = get_knmi_daily_rainfall_api(
                    url, stn, meteo_var, start, end, inseason, verbose)
            else:
                # daily data from meteorological stations
                url = 'http://projects.knmi.nl/klimatologie/daggegevens/getdata_dag.cgi'
                knmi_df, variables, stations = get_knmi_daily_meteo_api(
                    url, stn, meteo_var, start, end, inseason, verbose)
        else:
            if interval.startswith('hour'):
                # hourly data from meteorological stations
                raise NotImplementedError()
            elif meteo_var == 'RD':
                # daily data from rainfall-stations
                if stn_name is None:
                    stations_df = get_stations(meteo_var=meteo_var)
                    stn_name = get_station_name(int(stn), stations_df)
                knmi_df, variables = get_knmi_daily_rainfall_url(
                    stn, stn_name, meteo_var, start, end, inseason, verbose)
            else:
                # daily data from meteorological stations
                knmi_df, variables, stations = get_knmi_daily_meteo_url(
                    stn, meteo_var, start, end, inseason, verbose)
    except (ValueError, KeyError) as e:
        if verbose:
            print(e)
        if raise_exceptions:
            raise ValueError(e)

    return knmi_df, variables, stations


def get_knmi_daily_rainfall_api(url, stn, meteo_var,
                                start=None, end=None, inseason=False,
                                verbose=False):
    """download and read knmi daily rainfall.

    Parameters
    ----------
    url : str
        download url.
    stn : str
        station number.
    meteo_var : str
        must be 'RD'.
    start : pd.TimeStamp
        start time of observations.
    end : pd.TimeStamp
        end time of observations.
    inseason : boolean
        flag to obtain inseason data.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Raises
    ------
    ValueError
        if there is no data for the provided stn an error is raised.

    Returns
    -------
    pd.DataFrame
        measurements.
    variables : dictionary
        additional information about the variables
    """

    data = {
        'inseason': str(int(inseason)),
        'vars': meteo_var,
        'stns': stn}

    if start is not None:
        data['start'] = start.strftime('%Y%m%d')

    if end is not None:
        data['end'] = end.strftime('%Y%m%d')

    result = requests.get(url, params=data).text

    f = StringIO(result)
    knmi_df, variables = read_knmi_daily_rainfall(
        f, meteo_var, verbose=verbose)
    if int(stn) not in knmi_df.STN.unique():
        raise ValueError(f'KNMI station {stn} does not exists or has no '
                         'measurements in the given period')

    return knmi_df[[meteo_var]], variables


def get_knmi_daily_rainfall_url(stn, stn_name, meteo_var,
                                start=None, end=None, inseason=False,
                                use_cache=True,
                                verbose=False):
    """download and read knmi daily rainfall.

    Parameters
    ----------
    stn : str
        station number.
    stn_name : str
        the name of the station in capital letters, can be tricky
    meteo_var : str
        must be 'RD'.
    start : pd.TimeStamp
        start time of observations.
    end : pd.TimeStamp
        end time of observations.
    inseason : boolean
        flag to obtain inseason data.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Raises
    ------
    ValueError
        if there is no data for the provided stn an error is raised.

    Returns
    -------
    pd.DataFrame
        measurements.
    variables : dictionary
        additional information about the variables
    """
    url = f"https://cdn.knmi.nl/knmi/map/page/klimatologie/gegevens/monv_reeksen/neerslaggeg_{stn_name}_{stn}.zip"

    # get file name
    basedir = os.path.join(tempfile.gettempdir(), 'knmi')
    if not os.path.isdir(basedir):
        os.mkdir(basedir)
    fname_zip = os.path.join(tempfile.gettempdir(),
                             'knmi', f'neerslaggeg_{stn}.zip')
    fname_dir = os.path.join(tempfile.gettempdir(),
                             'knmi', f'neerslaggeg_{stn}')
    fname_txt = os.path.join(fname_dir, f'neerslaggeg_{stn_name}_{stn}.txt')

    # check if file should be downloaded and unzipped
    donwload_and_unzip = True
    if os.path.exists(fname_txt) and use_cache:
        fname_time_mod = dt.datetime.fromtimestamp(os.path.getmtime(fname_txt))
        if fname_time_mod > end:
            donwload_and_unzip = False

    if donwload_and_unzip:
        # download zip file
        r = requests.get(url, stream=True)
        if r.status_code != 200:
            raise ValueError(
                f'invalid url {url} please check station name {stn_name}')
        with open(fname_zip, 'wb') as fd:
            for chunk in r.iter_content(chunk_size=128):
                fd.write(chunk)

        # unzip file
        util.unzip_file(fname_zip, fname_dir, force=True,
                        preserve_datetime=True)

    with open(fname_txt, 'r') as f:
        line = f.readline()
        # get meteo var
        for iline in range(50):
            if 'RD   ' in line:  # RD komt te vaak voor vandaar de spaties
                key, item = line.split('=')
                variables = {key.strip(): item.strip()}
                break
            line = f.readline()

        # get dataframe
        for iline in range(50):
            if 'STN' in line:
                columns = line.strip('# ').strip('\n').split(',')
                columns = [x.strip(' ') for x in columns]
                values = f.readline()
                if values == '\n':
                    values = f.readline()
            line = f.readline()

        df = pd.read_csv(f, header=None, names=columns, na_values='     ')

        df.set_index(pd.to_datetime(df.YYYYMMDD, format='%Y%m%d'),
                     inplace=True)
        df = df.drop('YYYYMMDD', axis=1)

        if df.index.duplicated().sum() > 0:
            df = df.loc[~df.index.duplicated(keep='first')]
            if verbose:
                print('duplicate indices removed from RD measurements')

        # sometimes the last row is empty, check for that and remove it
        if not df.empty:
            if df.iloc[-1].isna().any():
                if verbose:
                    print('last row contains no data, remove last row')
                df = df.drop(index=df.index[-1])
                df.loc[:, meteo_var] = df[meteo_var].astype(float)

        # daily precipitation amount in 0.1 mm over the period 08.00
        # preceding day - 08.00 UTC present day
        df.index = df.index + pd.to_timedelta(8, unit='h')

        # from UT to UT+1 (standard-time in the Netherlands)
        df.index = df.index + pd.to_timedelta(1, unit='h')

        df, variables = _transform_variables(df, variables, verbose)

    return df.loc[start:end, [meteo_var]], variables


def _read_knmi_header(f, verbose=False):

    variables = {}
    line = f.readline()
    if 'DOCTYPE HTML PUBLIC' in line:
        if verbose:

            print(f.read())
        raise ValueError('Internal Server Error')
    for iline in range(500):
        if ' = ' in line:
            line = line.lstrip(' #').strip('\n')
            varDes = line.split(' = ')
            variables[varDes[0].strip()] = varDes[1].strip()

        if 'STN,YY' in line:
            header = line.split(',')
            header = [item.lstrip().rstrip() for item in header]
            break

        line = f.readline()

    if iline > 498:
        raise ValueError('cannot read measurements from file')

    return f, variables, header


def _transform_variables(df, variables, verbose=False):

    for key, value in variables.items():
        # test if key existst in data
        if key not in df.keys():
            if key == 'YYYYMMDD' or key == 'HH':
                pass
            elif key == 'T10N':
                variables.pop(key)
                key = 'T10'
            else:
                raise NameError(key + ' does not exist in data')

        if '0.1 ' in value:
            if verbose:
                print(f'transform {key}, {value} from 0.1 to 1')
            # transform 0.1 to 1
            df[key] = df[key] * 0.1
            value = value.replace('0.1 ', '')
        if ' tiende ' in value:
            if verbose:
                print(f'transform {key}, {value} from 0.1 to 1')
            # transform 0.1 to 1
            df[key] = df[key] * 0.1
            value = value.replace(' tiende ', ' ')
        if ' mm' in value:
            if verbose:
                print(f'transform {key}, {value} from mm to m')
            # transform mm to m
            df[key] = df[key] * 0.001
            value = value.replace(' mm', ' m')
        if ' millimeters' in value:
            if verbose:
                print(f'transform {key}, {value} from mm to m')
            # transform mm to m
            df[key] = df[key] * 0.001
            value = value.replace(' millimeters', ' m')
        # Store new variable
        variables[key] = value

    return df, variables


def read_knmi_daily_rainfall(f, meteo_var, verbose=False):

    f, variables, header = _read_knmi_header(f)

    df = pd.read_csv(f, header=None, names=header, na_values='     ')
    f.close()

    df.set_index(pd.to_datetime(df.YYYYMMDD, format='%Y%m%d'),
                 inplace=True)
    df = df.drop('YYYYMMDD', axis=1)

    if df.index.duplicated().sum() > 0:
        df = df.loc[~df.index.duplicated(keep='first')]
        if verbose:
            print('duplicate indices removed from RD measurements')

    # sometimes the last row is messed up, check for that and remove it
    if not df.empty:
        if df.iloc[-1].isna().any():
            if verbose:
                print('last row contains no data, remove last row')
            df = df.drop(index=df.index[-1])
            df.loc[:, meteo_var] = df[meteo_var].astype(float)

    # daily precipitation amount in 0.1 mm over the period 08.00
    # preceding day - 08.00 UTC present day
    df.index = df.index + pd.to_timedelta(8, unit='h')

    # from UT to UT+1 (standard-time in the Netherlands)
    df.index = df.index + pd.to_timedelta(1, unit='h')

    df, variables = _transform_variables(df, variables, verbose)

    return df, variables


def _read_station_location(f, verbose=False):

    stations = None

    line = f.readline()
    for iline in range(30):
        if 'STN' in line:
            titels = line.strip('# ').split()
            titels = [x.replace('(', '_') for x in titels]
            titels = [x.replace(r')', '') for x in titels]

            values = f.readline().strip('# ').strip().replace(':', '')
            values = re.split(r'\s{2,}', values)

            # Create pd.DataFrame for station data
            stations = pd.DataFrame(columns=titels, data=[values])
            stations.set_index(['STN'], inplace=True)
            for col in stations.columns:
                try:
                    stations.loc[:, col] = stations[col].astype(float)
                except ValueError:
                    pass

            if ':' in f.readline():
                raise ValueError(
                    'KNMI station number not recognized please provide '
                    'valid meteo station number')
            break

        line = f.readline()

    if stations is None:
        if verbose:
            print('could not find stations')

    return f, stations


def get_knmi_daily_meteo_api(url, stn, meteo_var, start, end, inseason, verbose=False):
    """download and read knmi daily meteo data.

    Parameters
    ----------
    url : str
        download url.
    stn : str
        station number.
    meteo_var : str
        e.g. 'EV24'.
    start : pd.TimeStamp
        start time of observations.
    end : pd.TimeStamp
        end time of observations.
    inseason : boolean
        flag to obtain inseason data.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    pd.DataFrame
        measurements.
    variables : dictionary
        additional information about the variables
    stations : pd.DataFrame
        additional data about the measurement station
    """
    data = {
        'start': start.strftime('%Y%m%d'),
        'end': end.strftime('%Y%m%d'),
        'inseason': str(int(inseason)),
        'vars': meteo_var,
        'stns': stn,
    }

    result = requests.get(url, params=data).text

    f = StringIO(result)
    knmi_df, variables, stations = read_knmi_daily_meteo(f, verbose=verbose)

    return knmi_df[[meteo_var]], variables, stations


def get_knmi_daily_meteo_url(stn, meteo_var, start, end,
                             use_cache=True,
                             verbose=False):
    """download and read knmi daily meteo data.

    Parameters
    ----------
    url : str
        download url.
    stn : str
        station number.
    meteo_var : str
        e.g. 'EV24'.
    start : pd.TimeStamp
        start time of observations.
    end : pd.TimeStamp
        end time of observations.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    pd.DataFrame
        measurements.
    variables : dictionary
        additional information about the variables
    stations : pd.DataFrame
        additional data about the measurement station
    """
    url = f"https://cdn.knmi.nl/knmi/map/page/klimatologie/gegevens/daggegevens/etmgeg_{stn}.zip"

    # get file name
    basedir = os.path.join(tempfile.gettempdir(), 'knmi')
    if not os.path.isdir(basedir):
        os.mkdir(basedir)
    fname_zip = os.path.join(tempfile.gettempdir(),
                             'knmi', f'etmgeg_{stn}.zip')
    fname_dir = os.path.join(tempfile.gettempdir(), 'knmi', f'etmgeg_{stn}')
    fname_txt = os.path.join(fname_dir, f'etmgeg_{stn}.txt')

    # check if file should be downloaded and unzipped
    donwload_and_unzip = True
    if os.path.exists(fname_txt) and use_cache:
        fname_time_mod = dt.datetime.fromtimestamp(os.path.getmtime(fname_txt))
        if fname_time_mod > end:
            donwload_and_unzip = False

    if donwload_and_unzip:
        # download zip file
        r = requests.get(url, stream=True)
        with open(fname_zip, 'wb') as fd:
            for chunk in r.iter_content(chunk_size=128):
                fd.write(chunk)

        # unzip file
        util.unzip_file(fname_zip, fname_dir, force=True,
                        preserve_datetime=True)

    variables = None
    with open(fname_txt, 'r') as f:
        line = f.readline()
        # get meteo var
        for iline in range(50):
            if meteo_var in line:
                key, item = line.split('=')
                variables = {key.strip(): item.strip()}
                break
            line = f.readline()

        if variables is None:
            raise ValueError(f'could not find {meteo_var} for station {stn}')

        # get dataframe
        for iline in range(50):
            if 'STN' in line:
                columns = line.strip('# ').strip('\n').split(',')
                columns = [x.strip(' ') for x in columns]
                values = f.readline()
                if values == '\n':
                    values = f.readline()
                df = pd.read_csv(f, header=None, names=columns,
                                 na_values='     ')
                df.set_index(pd.to_datetime(df.YYYYMMDD, format='%Y%m%d'),
                             inplace=True)
                df = df.drop('YYYYMMDD', axis=1)

                df = df.loc[df.index.notnull(), :]

                # add a full day for meteorological data, so that the
                # timestamp is at the end of the period in the data
                df.index = df.index + pd.to_timedelta(1, unit='d')

                # from UT to UT+1 (standard-time in the Netherlands)
                df.index = df.index + pd.to_timedelta(1, unit='h')
                df = df.loc[start:end, [meteo_var]]
                df = df.dropna()
                df, variables = _transform_variables(df, variables, verbose)
                break

            line = f.readline()

    return df, variables, None


def read_knmi_daily_meteo(f, verbose=False):

    f, stations = _read_station_location(f, verbose)
    f, variables, header = _read_knmi_header(f)
    header[0] = header[0].lstrip('# ')
    df = pd.read_csv(f, header=None, names=header, na_values='     ')
    f.close()

    df.set_index(pd.to_datetime(df.YYYYMMDD, format='%Y%m%d'),
                 inplace=True)
    df = df.drop('YYYYMMDD', axis=1)

    df = df.loc[df.index.notnull(), :]

    # add a full day for meteorological data, so that the
    # timestamp is at the end of the period in the data
    df.index = df.index + pd.to_timedelta(1, unit='d')

    # from UT to UT+1 (standard-time in the Netherlands)
    df.index = df.index + pd.to_timedelta(1, unit='h')

    df, variables = _transform_variables(df, variables, verbose)

    return df, variables, stations

   # return knmi_series


def get_knmi_hourly_api(url, stn, meteo_var, start, end):

    data = {
        'start': start.strftime('%Y%m%d') + '01',
        'end': end.strftime('%Y%m%d') + '24',
        'vars': meteo_var,
        'stns': stn,
    }

    result = requests.get(url, params=data).text
    f = StringIO(result)
    knmi_series = read_knmi_hourly(f)

    return knmi_series


def read_knmi_hourly(f):
    raise NotImplementedError('work in progress')
    knmi_df = pd.DataFrame()
    return knmi_df


def get_knmi_timeseries_xy(x, y, meteo_var, start, end,
                           stn_name=None,
                           fill_missing_obs=True,
                           interval='daily', inseason=False,
                           raise_exceptions=False,
                           verbose=False):
    """Get timeseries with measurements from station closest to the given (x,y)
    coördinates.

    Parameters
    ----------
    x : int or float
        x coördinate in m RD.
    y : int or float
        y coördinate in m RD.
    meteo_var : str
        e.g. 'EV24'.
    start : pd.TimeStamp
        start time of observations.
    end : pd.TimeStamp
        end time of observations.
    fill_missing_obs : bool
        if True missing observations are filled with values of next closest
        KNMI station
    interval : str, optional
        desired time interval for observations. The default is 'daily'.
    inseason : boolean, optional
        flag to obtain inseason data. The default is False
    raise_exceptions : bool, optional
        if True you get errors when no data is returned. The default is False.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    knmi_df : pd.DataFrame
        time series.
    meta : dic
        dictionary with metadata.
    """

    # get station
    stations = get_stations(meteo_var=meteo_var)
    stn = get_nearest_stations_xy(x, y, meteo_var, stations=stations)[0]

    # download data
    if fill_missing_obs:
        knmi_df, variables, station_meta = \
            fill_missing_measurements(stn, stn_name, meteo_var, start, end,
                                      interval, raise_exceptions,
                                      verbose=verbose)
    else:
        knmi_df, variables, station_meta = \
            download_knmi_data(stn, stn_name, meteo_var, start, end,
                               interval, inseason, raise_exceptions,
                               verbose=verbose)

    if not station_meta is None:
        meta = station_meta.to_dict()
    else:
        meta = {}
    meta.update(variables)

    # set metadata
    name = meteo_var + ' ' + stations.loc[stn, 'naam']
    x = stations.loc[stn, 'x']
    y = stations.loc[stn, 'y']
    meta.update({'x': x, 'y': y, 'station': stn, 'name': name})

    return knmi_df, meta


def get_knmi_timeseries_stn(stn, meteo_var, start, end,
                            fill_missing_obs=True, interval='daily',
                            inseason=False, raise_exceptions=False,
                            verbose=False):
    """Get a knmi time series and metadata.

    Parameters
    ----------
    stn : int or str
        measurement station e.g. 829.
    meteo_var : str, optional
        observation type e.g. "RD" or "EV24".
    start : str, datetime or None, optional
        start date of observations. The default is None.
    end : str, datetime or None, optional
        end date of observations. The default is None.
    fill_missing_obs : bool, optional
        if True nan values in time series are filled with nearby time series.
        The default is True.
    interval : str, optional
        desired time interval for observations. The default is 'daily'.
    inseason : boolean, optional
        flag to obtain inseason data. The default is False
    raise_exceptions : bool, optional
        if True you get errors when no data is returned. The default is True.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    knmi_df : pandas.DataFrame
        time series with measurements.
    meta : dictionary
        metadata from the measurement station.
    """

    # get station
    stations = get_stations(meteo_var=meteo_var)
    stn_name = get_station_name(stn, stations)

    # download data
    if fill_missing_obs:
        knmi_df, variables, station_meta = \
            fill_missing_measurements(stn, stn_name, meteo_var, start, end,
                                      interval, raise_exceptions,
                                      verbose=verbose)
    else:
        knmi_df, variables, station_meta = \
            download_knmi_data(stn, stn_name, meteo_var, start, end,
                               interval, inseason, use_api=False,
                               raise_exceptions=raise_exceptions,
                               verbose=verbose)
    if not station_meta is None:
        meta = station_meta.to_dict()
    else:
        meta = {}
    meta.update(variables)

    # set metadata
    name = meteo_var + ' ' + stations.loc[stn, 'naam']
    x = stations.loc[stn, 'x']
    y = stations.loc[stn, 'y']
    meta.update({'x': x, 'y': y, 'station': stn, 'name': name})

    return knmi_df, meta


def get_knmi_obslist(locations=None, stns=None, xmid=None, ymid=None,
                     meteo_vars=["RD"], start=[None], end=[None],
                     ObsClass=None, fill_missing_obs=True,
                     normalize_index=True, interval='daily', inseason=False,
                     cache=False, raise_exceptions=False, verbose=False):
    """Get a list of observations of knmi stations. Either specify a list of
    knmi stations (stns) or a dataframe with x, y coordinates (locations).

    Parameters
    ----------
    locations : pd.DataFrame or None
        dataframe with x and y coordinates. The default is None
    stns : list of str or None
        list of knmi stations. The default is None
    xmid : np.array, optional
        x coördinates of the cell centers of your grid shape(ncol)
    ymid : np.array, optional
        y coördinates of the cell centers of your grid shape(nrow)
    meteo_vars : list or tuple of str
        meteo variables e.g. ["RD", "EV24"]. The default is ("RD")
    start : list of str, datetime or None]
        start date of observations per meteo variable. The default is [None]
    end : list of str, datetime or None]
        end date of observations per meteo variable. The default is [None]
    ObsClass : type or None
        class of the observations, only KnmiObs is supported for now. The
        default is None
    fill_missing_obs : bool, optional
        if True nan values in time series are filled with nearby time series.
        The default is True.
    normalize_index : bool, optional
        if True the index is normalized.
    interval : str, optional
        desired time interval for observations. The default is 'daily'.
    inseason : boolean, optional
        flag to obtain inseason data. The default is False
    cache : boolean, optional
        if True the observation data will be cached or read from cache.
    raise_exceptions : bool, optional
        if True you get errors when no data is returned. The default is True.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    obs_list : list of obsevation objects
        collection of multiple point observations
    """
    obs_list = []
    for i, meteo_var in enumerate(meteo_vars):
        start[i], end[i] = _start_end_to_datetime(start[i], end[i])
        if stns is None:
            stations = get_stations(meteo_var=meteo_var)
            if (locations is None) and (xmid is not None):
                _stns = get_nearest_station_grid(xmid, ymid,
                                                 stations=stations,
                                                 meteo_var=meteo_var)
            elif locations is not None:
                _stns = get_nearest_station_df(locations,
                                               stations=stations,
                                               meteo_var=meteo_var)
            else:
                raise ValueError('stns, location and xmid are all None'
                                 'please specify one of these')
        else:
            _stns = stns

        for stn in _stns:
            if cache:
                cache_dir = os.path.join(tempfile.gettempdir(), 'knmi')
                if not os.path.isdir(cache_dir):
                    os.mkdir(cache_dir)

                fname = (f'{stn}-{meteo_var}-{start[i].strftime("%Y%m%d")}'
                         f'-{end[i].strftime("%Y%m%d")}'
                         f'-{fill_missing_obs}' + '.pklz')
                pklz_path = os.path.join(cache_dir, fname)

                if os.path.isfile(pklz_path):
                    if verbose:
                        print(f'reading {fname} from cache')
                    o = pd.read_pickle(pklz_path)
                else:
                    o = ObsClass.from_knmi(stn, meteo_var, start[i], end[i],
                                           fill_missing_obs=fill_missing_obs,
                                           interval=interval,
                                           inseason=inseason,
                                           raise_exceptions=raise_exceptions,
                                           verbose=verbose)
                    #o = o.loc[:, [meteo_var]]
                    o.to_pickle(pklz_path)
            else:
                o = ObsClass.from_knmi(stn, meteo_var, start[i], end[i],
                                       fill_missing_obs=fill_missing_obs,
                                       interval=interval,
                                       inseason=inseason,
                                       raise_exceptions=raise_exceptions,
                                       verbose=verbose)
                #o = o.loc[:, [meteo_var]]
            if normalize_index:
                o.index = o.index.normalize()

            obs_list.append(o)

    return obs_list


def add_missing_indices(knmi_df, stn, start, end, verbose=False):
    """when downloading KNMI data you don't always get a DataFrame with the
    periods that you provided in your request. Thus the index does not cover
    the complete period that you are interested in. This function adds the
    missing period to the index of the DataFrame.

    Parameters
    ----------
    knmi_df : pd.DataFrame
        data from one station from one type of observation, with additional
        column to see which station is used to fill the value
    stn : int or str
        measurement station.
    start : pd.TimeStamp
        start time of observations.
    end : pd.TimeStamp
        end time of observations.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    knmi_df : pd.DataFrame
        data from one station from one type of observation
    """
    # check if given dates are more or less similar than measurement dates
    if (knmi_df.index[0] - start).days < 2:
        new_start = knmi_df.index[0]
    else:
        new_start = pd.Timestamp(year=start.year, month=start.month,
                                 day=start.day, hour=knmi_df.index[0].hour,
                                 minute=knmi_df.index[0].minute,
                                 second=knmi_df.index[0].second)
        if verbose:
            print(
                f'station {stn} has no measurements before {knmi_df.index[0]}')

    if (end - knmi_df.index[-1]).days < 2:
        new_end = knmi_df.index[-1]
    else:
        new_end = pd.Timestamp(year=end.year, month=end.month, day=end.day,
                               hour=knmi_df.index[-1].hour,
                               minute=knmi_df.index[-1].minute,
                               second=knmi_df.index[-1].second)
        if verbose:
            print(
                f'station {stn} has no measurements after {knmi_df.index[-1]}')

    # add missing indices
    new_index = pd.date_range(new_start, new_end, freq='D')
    knmi_df = knmi_df.reindex(new_index)

    return knmi_df


def fill_missing_measurements(stn, stn_name=None, meteo_var='RD',
                              start=None, end=None,
                              interval='daily', raise_exceptions=False,
                              verbose=False):
    """fill missing measurements in knmi data.

    Parameters
    ----------
    stn : int or str
        measurement station.

    interval : str, optional
        desired time interval for observations. The default is 'daily'.
    meteo_var : str, optional
        observation type. The default is 'RD'.
    start : str, datetime or None, optional
        start date of observations. The default is None.
    end : str, datetime or None, optional
        end date of observations. The default is None.
    raise_exceptions : bool, optional
        if True you get errors when no data is returned. The default is True.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    knmi_df : pd.DataFrame
        data from one station from one type of observation, with additional
        column to see which station is used to fill the value
    variables : dictionary
        information about the observerd variables
    stations : pd.DataFrame
        information about the measurement station.
    """

    if not isinstance(meteo_var, str):
        raise (TypeError(f'meteo var should be string not {type(meteo_var)}'))
    # get the location of the stations
    stations = get_stations(meteo_var=meteo_var)

    start, end = _start_end_to_datetime(start, end)
    if (meteo_var == 'RD') and (end > (dt.datetime.now() - pd.Timedelta(90, unit='D'))):
        end = min(end, _check_latest_measurement_date_RD_debilt(verbose=verbose))

    knmi_df, variables, station_meta = \
        download_knmi_data(stn, stn_name, meteo_var, start=start,
                           end=end, interval=interval,
                           inseason=False,
                           raise_exceptions=raise_exceptions,
                           verbose=verbose)
    if verbose:
        print('Download ' + meteo_var + ' from ' +
              str(stn) + ' ' + stations.loc[stn, 'naam'])

    # if the first station cannot be read, read another station as the first
    ignore = [stn]
    while knmi_df.empty:
        stn = get_nearest_station_df(
            stations.loc[[stn]], meteo_var=meteo_var, ignore=ignore)[0]
        stn_name = get_station_name(stn, stations)
        if verbose:
            print('Download ' + meteo_var + ' from ' +
                  str(stn) + ' ' + stations.loc[stn, 'naam'])
        knmi_df, variables, station_meta = \
            download_knmi_data(stn, stn_name, meteo_var, start=start,
                               end=end, interval=interval,
                               inseason=False,
                               raise_exceptions=raise_exceptions,
                               verbose=verbose)
        ignore.append(stn)

    # find missing values
    knmi_df = add_missing_indices(knmi_df, stn, start, end, verbose)

    missing = knmi_df[meteo_var].isna()
    if verbose:
        print(f'station {stn} has {missing.sum()} missing measurements')

    # fill missing values
    while np.any(missing) and not np.all(missing):
        stn_comp = get_nearest_station_df(
            stations.loc[[stn]], meteo_var=meteo_var, ignore=ignore)

        if verbose:
            print(f'trying to fill {missing.sum()} '
                  f'measurements with station {stn_comp}')
        if stn_comp is None:
            if verbose:
                print('could not fill all missing measurements there are '
                      'no stations left to check')
            missing[:] = False
            break
        else:
            stn_comp = stn_comp[0]
        stn_name_comp = get_station_name(stn_comp, stations)
        knmi_df_comp, _, __ = \
            download_knmi_data(stn_comp, stn_name_comp, meteo_var,
                               start=start, end=end,
                               interval=interval,
                               inseason=False,
                               raise_exceptions=raise_exceptions,
                               verbose=verbose)

        if knmi_df_comp.empty:
            if verbose:
                print(f'station {stn_comp} cannot be downloaded')
        else:
            # dropnans from new data
            knmi_df_comp = knmi_df_comp.loc[~knmi_df_comp[meteo_var].isna(), :]
            # get index of missing data in original timeseries
            missing_idx = missing.loc[missing].index
            # if any missing are in the new data, update
            if missing_idx.isin(knmi_df_comp.index).any():
                # index for missing but in newly downloaded data
                ix_idx = missing_idx.intersection(knmi_df_comp.index)
                # update missing data
                knmi_df.loc[ix_idx,
                            meteo_var] = knmi_df_comp.loc[ix_idx, meteo_var]
                # add source station number
                knmi_df.loc[ix_idx, 'station_opvulwaarde'] = str(stn_comp)

        missing = knmi_df[meteo_var].isna()
        ignore.append(stn_comp)

    return knmi_df, variables, station_meta



# -*- coding: utf-8 -*-
"""Created on Wed Sep 12 12:15:42 2018.

@author: Artesia
"""
import warnings

import pandas as pd

from .. import observation as obs


def _fieldlogger_groundwater_settings(df, verbose=False):
    """Get default settings for groundwater observations.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe with observation point data

    Returns
    -------
    heading
    name
    subname
    inputfield
    properties
    group
    """

    if verbose:
        print('use default settings for groundwater observations')

    heading = ('use_standard_time;NO\nNAME;INPUTTYPE;HINT'
               '\nStand;number;GWS t.o.v. bkpb (cm)'
               '\nOpmerking;text;maak opmerking...\n')
    name = df.locatie
    subname = df.index
    inputfield = 'Stand|Opmerking'
    properties = df.obs.apply(
        lambda o: "bovenkant peilbuis|{:+.2f} mNAP|"
        "onderkant peilbuis|{:+.2f} mNAP|"
        "Maximaal gemeten|{:+.2f} mNAP|"
        "Gemiddeldeld gemeten|{:+.2f} mNAP|"
        "Minimaal gemeten|{:+.2f} mNAP".format(
            o.meta['bovenkant_filter'],
            o.meta['onderkant_filter'],
            o['stand_m_tov_nap'].max(),
            o['stand_m_tov_nap'].mean(),
            o['stand_m_tov_nap'].min()))
    group = df.name

    return heading, name, subname, inputfield, properties, group


def _fieldlogger_waterlevel_settings(df, verbose=False):
    """Get default  settings for waterlevel observations.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe with observation point data

    Returns
    -------
    name
    subname
    inputfield
    properties
    group
    """

    if verbose:
        print('use default settings for waterlevel observations')

    heading = ('use_standard_time;NO\nNAME;INPUTTYPE;HINT\n'
               'Stand;number;oppw t.o.v. kruinhoogte stuw (cm)\n'
               'Opmerking;text;maak opmerking...\n')
    name = df.index
    subname = df.index
    inputfield = 'Stand|Opmerking'
    properties = df.obs.apply(
        lambda o: "Maximaal gemeten|{:+.2f} mNAP|"
        "Gemiddeldeld gemeten|{:+.2f} mNAP|"
        "Minimaal gemeten|{:+.2f} mNAP".format(
            o['stand_m_tov_nap'].max(),
            o['stand_m_tov_nap'].mean(),
            o['stand_m_tov_nap'].min()))
    group = df.name

    return heading, name, subname, inputfield, properties, group


def fieldlogger_csv_to_obs_list(fname, ObsClass=obs.GroundwaterObs):
    """Read a fieldlogger file into a list of observation objects.

    Parameters
    ----------
    fname : str
        name of the fieldlogger location csv file
    ObsClass : observation class
        type of fieldlogger observations

    Returns
    -------
    obs_list : list
        list of observation objects
    """
    # read header
    fieldlogger_meta = {}
    with open(fname, 'r') as f:
        lcounter = 0
        line = f.readline()
        fieldlogger_meta[line.split(';')[0]] = line.strip().split(';')[1]
        while len(line.split(';')) < 4:
            line = f.readline()
            if line == 'NAME;INPUTTYPE;HINT\n':
                columns = line.strip().split(';')
                inputfields_df = pd.DataFrame(columns=columns)
                line = f.readline()
                i = 0
                while line != 'GROUP;COLOR\n':
                    inputfields_df.loc[i] = line.strip().split(';')
                    line = f.readline()
                    i += 1
                fieldlogger_meta['inputfields'] = inputfields_df
                if line == 'GROUP;COLOR\n':
                    columns = line.strip().split(';')
                    groups_df = pd.DataFrame(columns=columns)
                    line = f.readline()
                    i = 0
                    while len(line.split(';')) < 3:
                        groups_df.loc[i] = line.strip().split(';')
                        line = f.readline()
                        i += 1
                    fieldlogger_meta['groups'] = groups_df
            lcounter += 1

        # read body
        df = pd.read_csv(f, sep=';', names=line.strip('\n').split(';'),
                         header=None, index_col='SUBNAME')

    # create obs objects
    obs_list = []
    for o in df.iterrows():
        name = o[0]
        meta = o[1].to_dict()

        # find coordinates
        if 'XCOOR' in meta and 'YCOOR' in meta:
            x = meta['XCOOR']
            y = meta['YCOOR']
        elif 'LAT' in meta and 'LON' in meta:
            warnings.warn('X and Y coordinates are set to lat/lon values!')
            x = meta['LON']
            y = meta['LAT']
        else:
            raise ValueError('could not find xy coordinates in .csv file')

        if 'PROPERTIES' in meta:
            properties = meta.pop('PROPERTIES').split('|')
            meta.update(dict(zip(properties[::2], properties[1::2])))

        obs_list.append(ObsClass(pd.DataFrame(), name=name,
                                 x=x, y=y,
                                 meta=meta))

    return obs_list, fieldlogger_meta


def df_to_fieldlogger_csv(df, fname, otype=None,
                          use_default_otype_settings=True,
                          heading=None,
                          name=None, subname=None, inputfield=None,
                          properties=None, group=None, group_color='blue',
                          verbose=False):
    """Write a csv file that can be read by the fieldlogger app.

    Notes
    -----
    1. option to change input fields for measurements is not yet added
    2. multiple groups are not yet supported

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe with observation point data
    fname : str
        name of the fieldlogger csv file
    otype : str, optional
        observation type, 'groundwater' and 'surfacewater' are supported
    use_default_otype_settings : boolean, optional
        use the default settings for this measurement type
    heading : str, optional
        heading of the csv, if None use default heading
    name : series or str, optional
        1st column in fieldlogger csv, if None use index of ObsCollection
    subname : series or str, optional
        2nd column in fieldlogger csv, if None use name + _sub
    inputfield : series or str, optional
        3th column in fieldlogger csv, if None use 'Stand|Opmerking'
    properties : series or str, optional
        4th column in fieldlogger csv, if None use empty string
    group : series or str, optional
        5th column in fieldlogger csv, if None do not add group to csv
    group_color : series or str, optional
        color of the group
    """

    # get default settings
    if otype == obs.GroundwaterObs and use_default_otype_settings:
        heading, name, subname, inputfield, properties, group = _fieldlogger_groundwater_settings(
            df, verbose)
    elif otype == obs.WaterlvlObs and use_default_otype_settings:
        heading, name, subname, inputfield, properties, group = _fieldlogger_waterlevel_settings(
            df, verbose)

    # csv heading
    if heading is None:
        heading = ('use_standard_time;NO\nNAME;INPUTTYPE;HINT\n'
                   'Stand;number;GWS t.o.v. bkpb (cm)\n'
                   'Opmerking;text;maak opmerking...\n')

    columns = ['SUBNAME', 'XCOOR', 'YCOOR', 'INPUTFIELD', 'PROPERTIES']

    # add group to heading
    if type(group) == str:
        heading += 'GROUP;COLOR\n{};{}\n'.format(group, group_color)
        columns.append('GROUP')
        if verbose:
            print('add data as group {}'.format(group))
    elif group is not None:
        raise NotImplementedError('multiple groups are not yet supported')
    else:
        if verbose:
            print('no group is specified')

    # csv body
    fieldlogger_df = df[['x', 'y']].copy()
    if name is not None:
        fieldlogger_df.index = name
    fieldlogger_df.index.name = 'NAME'

    if subname is None:
        fieldlogger_df['SUBNAME'] = df.index + '_sub'
    else:
        fieldlogger_df['SUBNAME'] = subname

    if inputfield is None:
        fieldlogger_df['INPUTFIELD'] = 'Stand|Opmerking'
    else:
        fieldlogger_df['INPUTFIELD'] = inputfield

    if properties is None:
        fieldlogger_df['PROPERTIES'] = ''
    else:
        fieldlogger_df['PROPERTIES'] = properties.values

    if group is not None:
        fieldlogger_df['GROUP'] = group

    fieldlogger_df.rename(columns={'x': 'XCOOR', 'y': 'YCOOR'}, inplace=True)
    fieldlogger_df.index.name = 'NAME'

    # write file
    with open(fname, 'w') as fo:
        fo.write(heading)
        fieldlogger_df.to_csv(fo, header=True, columns=columns, sep=';')

    return fieldlogger_df



# -*- coding: utf-8 -*-
"""Created on Wed Sep 12 12:15:42 2018.

@author: Artesia
"""
import pandas as pd
import pastastore as pst


def _get_metadata_from_obs(o, verbose=False):
    """internal method to get metadata in the right format for a pastas series.
    A pastas timeseries cannot handle the same metadata format as an
    observation object.

    Parameters
    ----------
    o : observations.Obs
        observation time series with metadata
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    meta : dictionary
        meta dictionary.
    """
    meta = dict()
    for attr_key in o._metadata:
        val = getattr(o, attr_key)
        if isinstance(val, (int, float, str, bool)):
            meta[attr_key] = val
        elif isinstance(val, dict):
            for k, v in val.items():
                if isinstance(v, (int, float, str, bool)):
                    meta[k] = v
                else:
                    if verbose:
                        print(
                            f'did not add {k} to metadata because datatype is {type(v)}')
        else:
            if verbose:
                print(
                    f'did not add {attr_key} to metadata because datatype is {type(val)}')
    return meta


def create_pastastore(oc,
                      pstore,
                      pstore_name='',
                      conn=None,
                      add_metadata=True,
                      obs_column='stand_m_tov_nap',
                      kind='oseries',
                      overwrite=False,
                      verbose=False):
    """add observations to a new or existing pastastore.

    Parameters
    ----------
    oc : observation.ObsCollection
        collection of observations
    pstore : pastastore.PastasProject, optional
        Existing pastastore, if None a new project is created
    pstore_name : str, optional
        Name of the pastastore only used if pstore is None
    conn : pastastore.connectors
        connector for database
    obs_column : str, optional
        Name of the column in the Obs dataframe to be used
    kind : str, optional
        The kind of series that is added to the pastas project
    add_metadata : boolean, optional
        If True metadata from the observations added to the project.
    overwrite: bool, optional
        Whether to overwrite existing data in store.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    pstore : pastastore.PastasProject
        the pastas project with the series from the ObsCollection
    """
    if pstore is None:
        if conn is None:
            conn = pst.DictConnector("my_conn")
        pstore = pst.PastaStore(pstore_name, connector=conn)

    for o in oc.obs.values:
        if verbose:
            print('add to pastastore -> {}'.format(o.name))

        if add_metadata:
            meta = _get_metadata_from_obs(o, verbose=verbose)
        else:
            meta = dict()

        if kind == 'oseries':
            pstore.conn.add_oseries(o[obs_column], o.name,
                                    metadata=meta,
                                    overwrite=overwrite)
        else:
            pstore.conn.add_stress(o[obs_column], o.name,
                                   kind, metadata=meta,
                                   overwrite=overwrite)

    return pstore


def create_pastas_project(oc, pr=None, project_name='',
                          obs_column='stand_m_tov_nap',
                          kind='oseries', add_metadata=True,
                          verbose=False):
    """add observations to a new or existing pastas project.

    Parameters
    ----------
    oc : observation.ObsCollection
        collection of observations
    pr : pastas.project, optional
        Existing pastas project, if None a new project is created
    project_name : str, optional
        Name of the pastas project only used if pr is None
    obs_column : str, optional
        Name of the column in the Obs dataframe to be used
    kind : str, optional
        The kind of series that is added to the pastas project
    add_metadata : boolean, optional
        If True metadata from the observations added to the project.
    verbose : boolean, optional
        Print additional information to the screen (default is False).

    Returns
    -------
    pr : pastas.project
        the pastas project with the series from the ObsCollection
    """
    import pastas as ps
    if pr is None:
        pr = ps.Project(project_name)

    for o in oc.obs.values:
        if verbose:
            print('add to pastas project -> {}'.format(o.name))

        if add_metadata:
            meta = _get_metadata_from_obs(o, verbose=verbose)
        else:
            meta = dict()

        series = ps.TimeSeries(o[obs_column], name=o.name, metadata=meta)
        pr.add_series(series, kind=kind)

    return pr


def read_project(pr, ObsClass, rename_dic={}):
    """Read pastas.Project into ObsCollection.

    Parameters
    ----------
    pr : pastas.Project
        Project to read
    ObsClass : Obs
        ObsClass to read data as, usually GroundwaterObs
    rename_dic : dict, optional
        rename columns in Project oseries dictionary, by default empty dict

    Returns
    -------
    list : list of Obs
        list of Obs containing oseries data
    """
    obs_list = []
    for index, row in pr.oseries.iterrows():
        metadata = row.to_dict()
        for key in rename_dic.keys():
            if key in metadata.keys():
                metadata[rename_dic[key]] = metadata.pop(key)

        s = pd.DataFrame(metadata.pop('series').series_original)
        s.rename(columns={index: 'stand_m_tov_nap'}, inplace=True)

        keys_o = ['name', 'x', 'y', 'locatie', 'filternr',
                  'metadata_available', 'maaiveld', 'meetpunt',
                  'bovenkant_filter', 'onderkant_filter']
        meta_o = {k: metadata[k] for k in keys_o if k in metadata}

        o = ObsClass(s, meta=metadata, **meta_o)
        obs_list.append(o)
    return obs_list



"""A pystore is a datastore for Pandas Dataframes designed to store timeseries.

The functions in this module aim to save an obs_collection to a pystore. The
main advantages of a pystore are:

- smaller file size compared to .csv files
- exchangable format, the pystore format is independent of the pc
  (unlike pickle)

A pystore with an ObsCollection has 3 layers:

1. directory with name of the pystore with all the information of an
   ObsCollection.
2. Inside the pystore directory are directories with the collections of the
   pystore. One pystore collection corresponds to a subcollection of an
   ObsCollection (which is an ObsCollection on its own).

   You can use pystore collections to group observations from an ObsCollection
   by a certain feature, for example the location of the observation. One
   location can have multiple observations.

3. Inside the pystore collection are directories with the items of the
   pystore. An item of a pystore contains all the data from an Observation
   object.
"""

import os

import pandas as pd
from tqdm import tqdm

try:
    import pystore
except ModuleNotFoundError:
    pass

from ..obs_collection import ObsCollection
from ..observation import GroundwaterObs


def set_pystore_path(pystore_path):
    """Set pystore path.

    Parameters
    ----------
    pystore_path : str
        path to location with stores
    """
    pystore.set_path(pystore_path)


def item_to_obs(item, ObsClass, nameby="item"):
    """convert pystore Item to ObsClass.

    Parameters
    ----------
    item : pystore.item.Item
        pystore Item
    ObsClass : type of Obs
        type of observation DataFrame, e.g. GroundwaterObs

    Returns
    -------
    ObsClass
        DataFrame containing observations
    """
    if len(item.data.index) == 0:
        df = pd.DataFrame(columns=item.data.columns)
    else:
        df = item.to_pandas()
    item.metadata["datastore"] = str(item.datastore)
    if nameby == "item":
        name = item.item
    elif nameby == "collection":
        name = item.collection
    elif nameby == "both":
        name = item.collection + "__" + item.item
    else:
        raise ValueError(
            "'{}' is not a valid option for 'nameby'".format(nameby))

    metadata = item.metadata
    if not "name" in metadata.keys():
        metadata["name"] = name
    else:
        metadata.update({"name": name})

    obs_attr_dic = {}
    for attr in ObsClass._metadata:
        if attr in metadata.keys():
            obs_attr_dic[attr] = metadata[attr]

    o = ObsClass(df, **obs_attr_dic, meta=metadata)
    return o


def collection_to_obslist(store, collection, ObsClass=GroundwaterObs,
                          item_names=None, nameby="item", verbose=True):
    """pystore collection to list of observations.

    Parameters
    ----------
    store : pystore.store
        pystore store
    collection : pystore.collection
        pystore collection
    ObsClass : type of Obs
        type of observation, by default GroundwaterObs
    item_names : list of str
        item (Observation) names that will be extracted from the store
        the other items (Observations) will be ignored.

    Returns
    -------
    list : list of ObsClass
        list of ObsClass DataFrames
    """
    obs_list = []
    if collection in store.collections:
        collection = store.collection(collection)
    else:
        if verbose:
            print("Not in store -> {0}".format(collection))
        return obs_list

    if item_names is None:
        items = collection.items
    else:
        items = set(item_names) & set(collection.items)

    for i in items:
        try:
            item = collection.item(i)
            o = item_to_obs(item, ObsClass, nameby=nameby)
            obs_list.append(o)
        except Exception:
            if verbose:
                print("Skipped -> {0}: {1}".format(collection.collection, i))
    return obs_list


def store_to_obslist(store, ObsClass=GroundwaterObs, collection_names=None,
                     item_names=None, nameby="item", verbose=True,
                     progressbar=False):
    """convert pystore to list of ObsClass.

    Parameters
    ----------
    store : pystore.store
        pystore store containing data
    ObsClass : type of Obs
        type of observation DataFrames, by default GroundwaterObs
    item_names : list of str
        item (Observation) names that will be extracted from the store
        the other items (Observations) will be ignored. if None all items
        are read.
    nameby : str
        pick whether obs are named by collection or item name

    Returns
    -------
    list : list of obs
        list of ObsClass DataFrames
    """
    store = pystore.store(store)
    obs_list = []
    if collection_names is None:
        collections = store.collections
    else:
        collections = collection_names
    for coll in (tqdm(collections) if progressbar else collections):
        obs_list += collection_to_obslist(store, coll,
                                          ObsClass=ObsClass,
                                          item_names=item_names,
                                          nameby=nameby, verbose=verbose)
    return obs_list


def read_store_metadata(store, items='all', verbose=False):
    """read only metadata from pystore.

    Parameters
    ----------
    store : pystore.store
        store containing data
    items : str, list of str, optional
        if 'all' read all items
        if 'first' read first item only
        if list of str, read those items
    verbose : bool, optional
        if True, print progress info

    Returns
    -------
    list : list of dictionaries
        list of dictionaries containing metadata
    """
    store = pystore.store(store)
    meta_list = []
    for coll in store.collections:
        c = store.collection(coll)
        if items == "all":
            item_list = c.list_items()
        elif items == 'first':
            item_list = list(c.list_items())[0:1]
        else:
            item_list = items
        for i in item_list:
            metadata = pystore.utils.read_metadata(c._item_path(i))
            if metadata is None:
                if verbose:
                    print("Cannot read metadata for {0}/{1}".format(coll, i))
                metadata = dict()
                metadata['item_name'] = ""
            else:
                metadata['item_name'] = i
            metadata['collection_name'] = coll
            meta_list.append(metadata)
    return meta_list


def pystore_obslist_to_obscollection(obs_list, name="obs_coll"):
    """convert list of Obs to ObsCollection.

    Parameters
    ----------
    obs_list : list of Obs
        list of Obs DataFrames
    name : str, optional
        name of the collection, by default "obs_coll"

    Returns
    -------
    ObsCollection :
        DataFrame containing all Obs
    """
    coldict = [o.to_collection_dict() for o in obs_list]
    obs_df = pd.DataFrame(coldict, columns=coldict[0].keys())
    obs_df.set_index('name', inplace=True)
    meta = {'fname': obs_list[0].meta["datastore"],
            'type': GroundwaterObs,
            'verbose': True}
    oc = ObsCollection(obs_df, name=name, meta=meta)
    return oc


def read_pystore(storename, pystore_path,
                 ObsClass, extent=None, collection_names=None,
                 item_names=None, nameby="item",
                 read_series=True, verbose=True, progressbar=False):
    # set path
    set_pystore_path(pystore_path)

    if not os.path.isdir(os.path.join(pystore_path, storename)):
        raise FileNotFoundError("pystore -> '{}' "
                                "does not exist".format(storename))

    # obtain item names within extent
    if extent is not None:
        meta_list = read_store_metadata(storename, items="first")
        obs_df = pd.DataFrame(meta_list)
        obs_df.set_index('item_name', inplace=True)
        obs_df['x'] = pd.to_numeric(obs_df.x, errors='coerce')
        obs_df['y'] = pd.to_numeric(obs_df.y, errors='coerce')
        item_names = obs_df[(obs_df.x > extent[0]) & (obs_df.x < extent[1]) & (
            obs_df.y > extent[2]) & (obs_df.y < extent[3])].index

    if read_series:
        obs_list = store_to_obslist(
            storename,
            ObsClass=ObsClass,
            collection_names=collection_names,
            item_names=item_names,
            nameby=nameby,
            verbose=verbose,
            progressbar=progressbar)

        return obs_list
    else:
        if item_names is None:
            item_names = "all"
        meta_list = read_store_metadata(storename,
                                        items=item_names,
                                        verbose=verbose)
        meta = {}
        obs_df = pd.DataFrame(meta_list)
        if nameby == "collection":
            obs_df.set_index('collection_name', inplace=True)
        elif nameby == "item":
            obs_df.set_index('item_name', inplace=True)
        elif nameby == "both":
            obs_df["name"] = obs_df["collection_name"] + "__" + \
                obs_df["item_name"]
            obs_df.set_index('name', inplace=True)
        else:
            raise ValueError("'{}' is not a valid option "
                             "for 'nameby'".format(nameby))
        return obs_df




import os
import zipfile

import numpy as np
import pandas as pd
from pyproj import Proj, transform
from tqdm import tqdm


def read_waterinfo_file(path_to_file, index_cols=None, return_metadata=False,
                        value_col=None, location_col=None,
                        xcol=None, ycol=None, transform_coords=True):
    """Read waterinfo file (CSV or zip)

    Parameters
    ----------
    path_to_file : str
        path to file

    Returns
    -------
    df : pandas.DataFrame
        DataFrame containing file content
    metadata : dict, optional
        dict containing metadata, returned if return_metadata is
        True, default is False
    """

    name = os.path.splitext(os.path.basename(path_to_file))[0]

    if path_to_file.endswith(".csv"):
        f = path_to_file
    elif path_to_file.endswith(".zip"):
        zf = zipfile.ZipFile(path_to_file)
        f = zf.open('{}.csv'.format(name))
    else:
        raise NotImplementedError("File type '{}' not supported!".format(
            os.path.splitext(path_to_file)[-1]))

    if index_cols is None:
        index_cols = ['WAARNEMINGDATUM', 'WAARNEMINGTIJD']

    if value_col is None:
        value_col = 'NUMERIEKEWAARDE'

    if location_col is None:
        location_col = "MEETPUNT_IDENTIFICATIE"

    if xcol is None:
        xcol = "X"

    if ycol is None:
        ycol = "Y"

    # read data
    df = pd.read_csv(f,
                     sep=';',
                     decimal=',',
                     encoding="ISO-8859-1",
                     parse_dates=[index_cols],
                     dayfirst=True,
                     infer_datetime_format=True,
                     index_col="_".join(index_cols))

    # do some conversions
    df.loc[df[value_col]
           == 999999999, value_col] = np.NaN
    df[value_col] = df[value_col] / 100.

    # parse metadata into dict
    if return_metadata:
        if len(df[location_col].unique()) > 1:
            raise ValueError("File contains data for more than one location!"
                             " Use ObsCollection.from_waterinfo()!")

        metadata = {}
        if transform_coords:
            x, y = transform(Proj('epsg:25831'),
                            Proj('epsg:28992'),
                            df[xcol].iloc[-1],
                            df[ycol].iloc[-1])
        else:
            x = df[xcol].iloc[-1] / 100.
            y = df[ycol].iloc[-1] / 100.
        metadata["name"] = df[location_col].iloc[-1]
        metadata["x"] = x
        metadata["y"] = y

        return df, metadata
    else:
        return df


def read_waterinfo_obs(file_or_dir, ObsClass, progressbar=False):
    """Read waterinfo file or directory and extract locations and observations.

    Parameters
    ----------
    file_or_dir : str
        path to file or directory
    ObsClass: Obs type
        type of Obs to store data in
    progressbar : bool, optional
        show progressbar if True, default is False

    Returns
    -------
    obs_collection : list
        list of Obs objects
    """

    # Waterinfo file
    if os.path.isfile(file_or_dir):
        files = [file_or_dir]
    # directory with waterinfo files (zips or csvs)
    elif os.path.isdir(file_or_dir):
        files = [os.path.join(file_or_dir, f)
                 for f in sorted(os.listdir(file_or_dir))]
    else:
        raise NotImplementedError("Provide path to file or directory!")

    # loop over files
    metadata = {}
    obs_collection = []
    for filenm in tqdm(files):
        # read file or zip
        df = read_waterinfo_file(filenm)

        # get location and convert to m RD
        for stn in df["MEETPUNT_IDENTIFICATIE"].unique():
            mask = df['MEETPUNT_IDENTIFICATIE'] == stn
            x, y = transform(Proj(init='epsg:25831'),
                             Proj(init='epsg:28992'),
                             df.loc[mask, 'X'][-1],
                             df.loc[mask, 'Y'][-1])
            metadata = {"name": stn, "x": x, "y": y}

            # add to list
            o = ObsClass(df.loc[mask, :], meta=metadata, **metadata)
            obs_collection.append(o)

    return obs_collection



"""Module with a number of observation classes.

The Obs class is a subclass of a pandas DataFrame with
additional attributes and methods. The specific classes (GroundwaterObs,
WaterlvlObs, ...) are subclasses of the Obs class.

The subclasses of a dataframe can have additional attributes and methods.
Additional attributes have to be defined in the '_metadata' attribute. In order
to keep the subclass methods and attributes when selecting or slicing an object
you need the '_constructor' method.


More information about subclassing pandas DataFrames can be found here:
http://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
"""

import warnings
import numpy as np
from pandas import DataFrame


class Obs(DataFrame):
    """class for point observations.

    An Obs object is a subclass of a pandas.DataFrame and allows for additional
    attributes and methods.
    More information about subclassing pandas DataFrames pandas can be
    found here:
    http://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas

    Parameters
    ----------
    x : int or float
        x coordinate of observation point
    y : int or float
        y coordinate of observation point
    name : str
        name
    meta : dictionary
        metadata
    filename : str
        filename with data of observation point
    """
    # temporary properties
    _internal_names = DataFrame._internal_names + ['none']
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ['x', 'y', 'name', 'meta', 'filename']

    def __init__(self, *args, **kwargs):
        """constructor of Obs class.

        *args must be input for the pandas.DataFrame constructor,
        **kwargs can be one of the attributes listed in _metadata or
        keyword arguments for the constructor of a pandas.DataFrame.
        """
        self.x = kwargs.pop('x', np.nan)
        self.y = kwargs.pop('y', np.nan)
        self.name = kwargs.pop('name', '')
        self.meta = kwargs.pop('meta', {})
        self.filename = kwargs.pop('filename', '')

        super(Obs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return Obs

    def to_collection_dict(self, include_meta=False):
        """get dictionary with registered attributes and their values of an Obs
        object.

        This method can be used to create a dataframe from a collection
        of Obs objects.

        Parameters
        ----------
        include_meta : boolean, optional
            include the meta dictionary in the collection dictionary,
            default is false

        Returns
        -------
        d : dictionary
            dictionary with Obs information
        """

        attrs = self._metadata.copy()
        if not include_meta:
            attrs.remove('meta')

        d = {}
        for att in attrs:
            d[att] = getattr(self, att)

        d['obs'] = self

        return d


class GroundwaterObs(Obs):
    """Class for groundwater quantity observations.

    Subclass of the Obs class. Has the following attributes:

    - locatie: 2 filters at one piezometer should have the same 'locatie'
    - filternr: 2 filters at one piezometer should have a different 'filternr'.
      a higher filter number is preferably deeper than a lower filter number.
    - bovenkant_filter: top op the filter in m NAP
    - onderkant_filter: bottom of the filter in m NAP
    - maaiveld: surface level in m NAP
    - meetpunt: ? in m NAP
    - metadata_available: boolean indicating if metadata is available for
      the measurement point.
    """

    _metadata = Obs._metadata + \
        ['locatie', 'filternr',
         'bovenkant_filter', 'onderkant_filter',
         'maaiveld', 'meetpunt', 'metadata_available'
         ]

    def __init__(self, *args, **kwargs):
        """
        *args must be input for the pandas.DataFrame constructor,
        **kwargs can be one of the attributes listed in _metadata or
        keyword arguments for the constructor of a pandas.DataFrame.

        if the pandas.DataFrame has a column 'stand_m_tov_nap' a lot of
        plotting and other methods will work automatically without changing
        the default arguments.
        """
        self.locatie = kwargs.pop('locatie', '')
        self.filternr = kwargs.pop('filternr', '')
        self.maaiveld = kwargs.pop('maaiveld', np.nan)
        self.meetpunt = kwargs.pop('meetpunt', np.nan)
        self.bovenkant_filter = kwargs.pop('bovenkant_filter', np.nan)
        self.onderkant_filter = kwargs.pop('onderkant_filter', np.nan)
        self.metadata_available = kwargs.pop('metadata_available', np.nan)

        super(GroundwaterObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return GroundwaterObs

    @classmethod
    def from_dino(cls, fname=None, location=None, filternr=1.,
                  tmin="1900-01-01", tmax="2040-01-01",
                  split_cluster=True,
                  **kwargs):
        """download dino data from the server.

        Parameters
        ----------
        fname : str, optional
            dino csv filename
        location : str, optional
            location of the peilbuis, i.e. B57F0077
        filternr : float, optional
            filter_nr of the peilbuis, i.e. 1.
        tmin : str
            start date in format YYYY-MM-DD
        tmax : str
            end date in format YYYY-MM-DD
        split_cluster : bool
            if False and the piezometer belongs to a cluster, the combined
            time series of the cluster is used. if True the indvidual time
            series of each piezometer is used. Default is True
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_csv if
            fname is not None and otherwise to io_dino.findMeetreeks
        """
        from .io import io_dino
        if fname is not None:
            # read dino csv file
            measurements, meta = io_dino.read_dino_groundwater_csv(
                fname, **kwargs)

            return cls(measurements, meta=meta, **meta)

        elif location is not None:
            measurements, meta = io_dino.download_dino_groundwater(location,
                                                                   filternr,
                                                                   tmin, tmax,
                                                                   split_cluster,
                                                                   **kwargs)

            if meta['metadata_available']:
                return cls(measurements, meta=meta,
                           x=meta.pop('x'), y=meta.pop('y'),
                           onderkant_filter=meta.pop('onderkant_filter'),
                           bovenkant_filter=meta.pop('bovenkant_filter'),
                           name=meta.pop('name'),
                           metadata_available=meta.pop('metadata_available'),
                           locatie=meta.pop('locatie'),
                           maaiveld=meta.pop('maaiveld'),
                           filternr=meta.pop('filternr'))
            else:
                return cls(measurements, meta=meta)
        else:
            raise ValueError(
                'specify fname or location to obtain groundwater heads')

    @classmethod
    def from_dino_server(cls, location, filternr=1.,
                         tmin="1900-01-01", tmax="2040-01-01",
                         **kwargs):
        """download dino data from the server.

        Parameters
        ----------
        location : str
            location of the peilbuis, i.e. B57F0077
        filternr : float
            filter_nr of the peilbuis, i.e. 1.0
        tmin : str
            start date in format YYYY-MM-DD
        tmax : str
            end date in format YYYY-MM-DD
        kwargs : key-word arguments
            these arguments are passed to dino.findMeetreeks functie
        """
        warnings.warn(
            "this method will be removed in future versions, use from_dino instead", DeprecationWarning)

        from .io import io_dino
        measurements, meta = io_dino.download_dino_groundwater(location,
                                                               filternr,
                                                               tmin, tmax,
                                                               **kwargs)

        if meta['metadata_available']:
            return cls(measurements, meta=meta,
                       x=meta.pop('x'), y=meta.pop('y'),
                       onderkant_filter=meta.pop('onderkant_filter'),
                       bovenkant_filter=meta.pop('bovenkant_filter'),
                       name=meta.pop('name'),
                       locatie=meta.pop('locatie'),
                       maaiveld=meta.pop('maaiveld'),
                       filternr=meta.pop('filternr'))
        else:
            return cls(measurements, meta=meta)

    @classmethod
    def from_dino_file(cls, fname=None, **kwargs):
        """read a dino csv file.

        Parameters
        ----------
        name : str, optional
            name of the peilbuis, i.e. B57F0077
        fname : str, optional
            dino csv filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_csv
        """
        warnings.warn(
            "this method will be removed in future versions, use from_dino instead", DeprecationWarning)

        from .io import io_dino
        if fname is not None:
            # read dino csv file
            measurements, meta = io_dino.read_dino_groundwater_csv(
                fname, **kwargs)

            return cls(measurements, meta=meta, **meta)
        else:
            raise ValueError(
                'specify either the name or the filename of the measurement point')

    @classmethod
    def from_artdino_file(cls, fname=None, **kwargs):
        """read a dino csv file (artdiver style).

        Parameters
        ----------
        name : str, optional
            name of the peilbuis, i.e. B57F0077
        fname : str, optional
            dino csv filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_csv
        """
        from .io import io_dino
        if fname is not None:
            # read dino csv file
            measurements, meta = io_dino.read_artdino_groundwater_csv(
                fname, **kwargs)

            return cls(measurements, meta=meta, **meta)
        else:
            raise ValueError('specify either the name or the filename of the '
                             'measurement point!')

    @classmethod
    def from_wiski(cls, fname, **kwargs):
        """[summary]

        Parameters
        ----------
        fname : [type]
            [description]

        Returns
        -------
        [type]
            [description]
        """
        from .io import io_wiski
        data, metadata = io_wiski.read_wiski_file(fname, **kwargs)

        return cls(data, meta=metadata, **metadata)

    @classmethod
    def from_pystore_item(cls, item):
        """Create GroundwaterObs DataFrame from Pystore item.

        Parameters
        ----------
        item : pystore.item.Item
            Pystore item

        Returns
        -------
        GroundwaterObs
            GroundwaterObs DataFrame
        """

        df = item.to_pandas()
        try:
            x = item.metadata["x"]
            y = item.metadata["y"]
        except KeyError:
            x = np.nan
            y = np.nan
        item.metadata["datastore"] = item.datastore
        return cls(df, x=x, y=y, meta=item.metadata)


class GroundwaterQualityObs(Obs):
    """class for groundwater quality (grondwatersamenstelling) point
    observations.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + \
        ['locatie', 'filternr', 'maaiveld', 'metadata_available']

    def __init__(self, *args, **kwargs):

        self.locatie = kwargs.pop('locatie', '')
        self.filternr = kwargs.pop('filternr', '')
        self.maaiveld = kwargs.pop('maaiveld', np.nan)
        self.metadata_available = kwargs.pop('metadata_available', np.nan)

        super(GroundwaterQualityObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return GroundwaterQualityObs

    @classmethod
    def from_dino(cls, fname, **kwargs):
        """read ad dino file with groundwater quality data.

        Parameters
        ----------
        fname : str
            dino txt filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_groundwater_quality_txt
        """
        from .io import io_dino
        measurements, meta = io_dino.read_dino_groundwater_quality_txt(
            fname, **kwargs)

        return cls(measurements, meta=meta, **meta)


class WaterlvlObs(Obs):
    """class for water level point observations.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ['locatie', 'metadata_available']

    def __init__(self, *args, **kwargs):

        self.locatie = kwargs.pop('locatie', '')
        self.metadata_available = kwargs.pop('metadata_available', np.nan)

        super(WaterlvlObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return WaterlvlObs

    @classmethod
    def from_dino(cls, fname, **kwargs):
        """read a dino file with waterlvl data.

        Parameters
        ----------
        fname : str
            dino csv filename
        kwargs : key-word arguments
            these arguments are passed to io_dino.read_dino_waterlvl_csv
        """
        from .io import io_dino
        measurements, meta = io_dino.read_dino_waterlvl_csv(fname, **kwargs)

        return cls(measurements, meta=meta, **meta)

    @classmethod
    def from_waterinfo(cls, fname, **kwargs):
        """Read data from waterinfo csv-file or zip.

        Parameters
        ----------
        fname : str
            path to file (file can zip or csv)

        Returns
        -------
        df : WaterlvlObs
            WaterlvlObs object

        Raises
        ------
        ValueError
            if file contains data for more than one location
        """
        from .io import io_waterinfo
        df, metadata = io_waterinfo.read_waterinfo_file(fname,
                                                        return_metadata=True,
                                                        **kwargs)
        return cls(df, meta=metadata, **metadata)


class ModelObs(Obs):
    """class for model point results.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ['model']

    def __init__(self, *args, **kwargs):

        self.model = kwargs.pop('model', '')

        super(ModelObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return ModelObs


class KnmiObs(Obs):
    """class for KNMI timeseries.

    Subclass of the Obs class
    """

    _metadata = Obs._metadata + ['station']

    def __init__(self, *args, **kwargs):

        self.station = kwargs.pop('station', np.nan)

        super(KnmiObs, self).__init__(*args, **kwargs)

    @property
    def _constructor(self):
        return KnmiObs

    @classmethod
    def from_knmi(cls, stn, variable, startdate=None, enddate=None,
                  fill_missing_obs=True, interval='daily', inseason=False,
                  raise_exceptions=True, verbose=False):
        """Get a KnmiObs object.

        Parameters
        ----------
        stn : int or str
            measurement station e.g. 829.
        variable : str, optional
            observation type e.g. "RD" or "EV24". See list with all variables
            below.
        startdate : str, datetime or None, optional
            start date of observations. The default is None.
        enddate : str, datetime or None, optional
            end date of observations. The default is None.
        fill_missing_obs : bool, optional
            if True nan values in time series are filled with nearby time series.
            The default is True.
        interval : str, optional
            desired time interval for observations. The default is 'daily'.
        inseason : boolean, optional
            flag to obtain inseason data. The default is False
        raise_exceptions : bool, optional
            if True you get errors when no data is returned. The default is False.
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        KnmiObs object with time series and attributes

        List of possible variables:
            DDVEC     = Vectorgemiddelde windrichting in graden (360=noord, 90=oost, 180=zuid, 270=west, 0=windstil/variabel). Zie http://www.knmi.nl/kennis-en-datacentrum/achtergrond/klimatologische-brochures-en-boeken / Vector mean wind direction in degrees (360=north, 90=east, 180=south, 270=west, 0=calm/variable)
            FHVEC     = Vectorgemiddelde windsnelheid (in 0.1 m/s). Zie http://www.knmi.nl/kennis-en-datacentrum/achtergrond/klimatologische-brochures-en-boeken / Vector mean windspeed (in 0.1 m/s)
            FG        = Etmaalgemiddelde windsnelheid (in 0.1 m/s) / Daily mean windspeed (in 0.1 m/s)
            FHX       = Hoogste uurgemiddelde windsnelheid (in 0.1 m/s) / Maximum hourly mean windspeed (in 0.1 m/s)
            FHXH      = Uurvak waarin FHX is gemeten / Hourly division in which FHX was measured
            FHN       = Laagste uurgemiddelde windsnelheid (in 0.1 m/s) / Minimum hourly mean windspeed (in 0.1 m/s)
            FHNH      = Uurvak waarin FHN is gemeten / Hourly division in which FHN was measured
            FXX       = Hoogste windstoot (in 0.1 m/s) / Maximum wind gust (in 0.1 m/s)
            FXXH      = Uurvak waarin FXX is gemeten / Hourly division in which FXX was measured
            TG        = Etmaalgemiddelde temperatuur (in 0.1 graden Celsius) / Daily mean temperature in (0.1 degrees Celsius)
            TN        = Minimum temperatuur (in 0.1 graden Celsius) / Minimum temperature (in 0.1 degrees Celsius)
            TNH       = Uurvak waarin TN is gemeten / Hourly division in which TN was measured
            TX        = Maximum temperatuur (in 0.1 graden Celsius) / Maximum temperature (in 0.1 degrees Celsius)
            TXH       = Uurvak waarin TX is gemeten / Hourly division in which TX was measured
            T10N      = Minimum temperatuur op 10 cm hoogte (in 0.1 graden Celsius) / Minimum temperature at 10 cm above surface (in 0.1 degrees Celsius)
            T10NH     = 6-uurs tijdvak waarin T10N is gemeten / 6-hourly division in which T10N was measured; 6=0-6 UT, 12=6-12 UT, 18=12-18 UT, 24=18-24 UT
            SQ        = Zonneschijnduur (in 0.1 uur) berekend uit de globale straling (-1 voor <0.05 uur) / Sunshine duration (in 0.1 hour) calculated from global radiation (-1 for <0.05 hour)
            SP        = Percentage van de langst mogelijke zonneschijnduur / Percentage of maximum potential sunshine duration
            Q         = Globale straling (in J/cm2) / Global radiation (in J/cm2)
            DR        = Duur van de neerslag (in 0.1 uur) / Precipitation duration (in 0.1 hour)
            RH        = Etmaalsom van de neerslag (in 0.1 mm) (-1 voor <0.05 mm) / Daily precipitation amount (in 0.1 mm) (-1 for <0.05 mm)
            RHX       = Hoogste uursom van de neerslag (in 0.1 mm) (-1 voor <0.05 mm) / Maximum hourly precipitation amount (in 0.1 mm) (-1 for <0.05 mm)
            RHXH      = Uurvak waarin RHX is gemeten / Hourly division in which RHX was measured
            PG        = Etmaalgemiddelde luchtdruk herleid tot zeeniveau (in 0.1 hPa) berekend uit 24 uurwaarden / Daily mean sea level pressure (in 0.1 hPa) calculated from 24 hourly values
            PX        = Hoogste uurwaarde van de luchtdruk herleid tot zeeniveau (in 0.1 hPa) / Maximum hourly sea level pressure (in 0.1 hPa)
            PXH       = Uurvak waarin PX is gemeten / Hourly division in which PX was measured
            PN        = Laagste uurwaarde van de luchtdruk herleid tot zeeniveau (in 0.1 hPa) / Minimum hourly sea level pressure (in 0.1 hPa)
            PNH       = Uurvak waarin PN is gemeten / Hourly division in which PN was measured
            VVN       = Minimum opgetreden zicht / Minimum visibility; 0: <100 m, 1:100-200 m, 2:200-300 m,..., 49:4900-5000 m, 50:5-6 km, 56:6-7 km, 57:7-8 km,..., 79:29-30 km, 80:30-35 km, 81:35-40 km,..., 89: >70 km)
            VVNH      = Uurvak waarin VVN is gemeten / Hourly division in which VVN was measured
            VVX       = Maximum opgetreden zicht / Maximum visibility; 0: <100 m, 1:100-200 m, 2:200-300 m,..., 49:4900-5000 m, 50:5-6 km, 56:6-7 km, 57:7-8 km,..., 79:29-30 km, 80:30-35 km, 81:35-40 km,..., 89: >70 km)
            VVXH      = Uurvak waarin VVX is gemeten / Hourly division in which VVX was measured
            NG        = Etmaalgemiddelde bewolking (bedekkingsgraad van de bovenlucht in achtsten, 9=bovenlucht onzichtbaar) / Mean daily cloud cover (in octants, 9=sky invisible)
            UG        = Etmaalgemiddelde relatieve vochtigheid (in procenten) / Daily mean relative atmospheric humidity (in percents)
            UX        = Maximale relatieve vochtigheid (in procenten) / Maximum relative atmospheric humidity (in percents)
            UXH       = Uurvak waarin UX is gemeten / Hourly division in which UX was measured
            UN        = Minimale relatieve vochtigheid (in procenten) / Minimum relative atmospheric humidity (in percents)
            UNH       = Uurvak waarin UN is gemeten / Hourly division in which UN was measured
            EV24      = Referentiegewasverdamping (Makkink) (in 0.1 mm) / Potential evapotranspiration (Makkink) (in 0.1 mm)
        """
        from .io import io_knmi

        ts, meta = io_knmi.get_knmi_timeseries_stn(
            stn, variable, startdate, enddate, fill_missing_obs,
            interval=interval, inseason=inseason,
            raise_exceptions=raise_exceptions,
            verbose=verbose
        )

        return cls(ts, meta=meta, station=meta['station'], x=meta['x'],
                   y=meta['y'], name=meta['name'])

    @classmethod
    def from_nearest_xy(cls, x, y, variable, startdate=None, enddate=None,
                        fill_missing_obs=True, interval='daily',
                        inseason=False, raise_exceptions=False, verbose=False):
        """Get KnmiObs object with measurements from station closest to the
        given (x,y) coördinates.

        Parameters
        ----------
        x : int or float
            x coördinate in m RD.
        y : int or float
            y coördinate in m RD.
        variable : str
            e.g. 'EV24'.
        startdate : str, datetime or None, optional
            start date of observations. The default is None.
        enddate : str, datetime or None, optional
            end date of observations. The default is None.
        fill_missing_obs : bool
            if True missing observations are filled with values of next closest
            KNMI station
        interval : str, optional
            desired time interval for observations. The default is 'daily'.
        inseason : boolean, optional
            flag to obtain inseason data. The default is False
        raise_exceptions : bool, optional
            if True you get errors when no data is returned. The default is False.
        verbose : boolean, optional
            Print additional information to the screen (default is False).

        Returns
        -------
        KnmiObs object with time series and attributes
        """
        from .io import io_knmi

        ts, meta = io_knmi.get_knmi_timeseries_xy(
            x, y, variable, startdate, enddate, fill_missing_obs,
            interval=interval, inseason=inseason,
            raise_exceptions=raise_exceptions,
            verbose=verbose
        )

        return cls(ts, meta=meta, station=meta['station'], x=meta['x'],
                   y=meta['y'], name=meta['name'])

    @classmethod
    def from_obs(cls, obs, variable, startdate=None, enddate=None,
                 fill_missing_obs=True, interval='daily', inseason=False,
                 raise_exceptions=False, verbose=False):

        from .io import io_knmi

        x = obs.x
        y = obs.y

        if startdate is None:
            startdate = obs.index[0]
        if enddate is None:
            enddate = obs.index[-1]

        ts, meta = io_knmi.get_knmi_timeseries_xy(
            x, y, variable, startdate, enddate, fill_missing_obs,
            interval=interval, inseason=inseason,
            raise_exceptions=raise_exceptions,
            verbose=verbose
        )

        return cls(ts, meta=meta, station=meta['station'], x=meta['x'],
                   y=meta['y'], name=meta['name'])



from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'readme.md'), encoding='utf-8') as f:
    l_d = f.read()

# Get the version.
version = {}
with open("hydropandas/version.py") as fp:
    exec(fp.read(), version)

setup(
    name='hydropandas',
    version=version['__version__'],
    description='hydropandas module by Artesia',
    long_description=l_d,
    long_description_content_type="text/markdown",
    url='https://github.com/ArtesiaWater/hydropandas',
    author='Artesia',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    platforms='Windows, Mac OS-X',
    install_requires=['numpy>=1.15',
                      'scipy>=1.2',
                      'matplotlib>=3.0',
                      'pandas>=0.24',
                      'tqdm >= 4.42.1'
                      'zeep>=3.4.0',
                      'requests >= 2.22.0',
                      'geopandas >= 0.6.0'
                      ],
    packages=find_packages(exclude=[]),
    package_data={"hydropandas": ["data/*.json"]},
    include_package_data=True
)



"""Copied from ./pandas/core/accessor.py."""
import warnings


class CachedAccessor:
    """Custom property-like object (descriptor) for caching accessors.

    Parameters
    ----------
    name : str
        The namespace this will be accessed under, e.g. ``df.foo``
    accessor : cls
        The class with the extension methods. The class' __init__ method
        should expect one of a ``Series``, ``DataFrame`` or ``Index`` as
        the single argument ``data``
    """

    def __init__(self, name, accessor):
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Dataset.geo
            return self._accessor
        accessor_obj = self._accessor(obj)
        # Replace the property with the accessor object. Inspired by:
        # http://www.pydanny.com/cached-property.html
        # We need to use object.__setattr__ because we overwrite __setattr__ on
        # NDFrame
        object.__setattr__(obj, self._name, accessor_obj)
        return accessor_obj


def _register_accessor(name, cls):
    def decorator(accessor):
        if hasattr(cls, name):
            warnings.warn(
                "registration of accessor {!r} under name {!r} for type "
                "{!r} is overriding a preexisting attribute with the same "
                "name.".format(accessor, name, cls),
                UserWarning,
                stacklevel=2,
            )
        setattr(cls, name, CachedAccessor(name, accessor))
        cls._accessors.add(name)
        return accessor

    return decorator


def register_obscollection_accessor(name):
    from .. import obs_collection as oc

    return _register_accessor(name, oc.ObsCollection)


def register_obs_accessor(name):
    from .. import observation as obs

    return _register_accessor(name, obs.Obs)



from .obs_collection import ObsCollection
from .observation import (Obs, GroundwaterObs, GroundwaterQualityObs,
                          KnmiObs, ModelObs, WaterlvlObs)
from .extensions import geo as _geo
from .extensions import gwobs as _gwobs
from .extensions import plots as _plots
from .extensions import stats as _stats
from .version import __version__
